@@ -2,13 +2,20 @@
 """
 
 import logging
+import mimetypes
 import re
+from io import BytesIO
 from pathlib import Path
+from time import sleep
 
+import geopandas as gpd
 import pandas as pd
-import pygsheets
 import pysftp
 import requests
+from googleapiclient.http import MediaIoBaseDownload
+from sqlalchemy import create_engine
+
+from . import utils
 
 logger = logging.getLogger(__name__)
 
@@ -16,9 +23,9 @@ logger = logging.getLogger(__name__)
 class GSheetLoader:
     """Loads data from a Google Sheets spreadsheet into a pandas data frame"""
 
-    def __init__(self, service_file):
-        self.gsheets_client = pygsheets.authorize(service_file=service_file)
+    def __init__(self, credentials):
         self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
+        self.gsheets_client = utils.authorize_pygsheets(credentials)
 
     def load_specific_worksheet_into_dataframe(self, sheet_id, worksheet, by_title=False):
         """Load a single worksheet from a spreadsheet into a dataframe by worksheet index or title
@@ -184,17 +191,21 @@ class GoogleDriveDownloader:
                 raise IndexError(f'Regex could not extract the file id from sharing link {sharing_link}') from err
         raise RuntimeError(f'Regex could not match sharing link {sharing_link}')
 
-    def download_file_from_google_drive(self, sharing_link, join_id):
+    def download_file_from_google_drive(self, sharing_link, join_id, pause=0.):
         """Download a publicly-shared image from Google Drive using it's sharing link
 
         Args:
             sharing_link (str): The publicly-shared link to the image.
             join_id (str or int): The unique key for the row (used for reporting)
+            pause (flt, optional): Pause the specified number of seconds before downloading. Defaults to 0.
 
         Returns:
             Path: Path of downloaded file or None if download fails/is not possible
         """
 
+        if pause:
+            self._class_logger.debug('Sleeping for %s', pause)
+        sleep(pause)
         if not sharing_link:
             self._class_logger.debug('Row %s has no attachment info', join_id)
             return None
@@ -213,19 +224,71 @@ class GoogleDriveDownloader:
             self._class_logger.warning(err)
             return None
 
-    #: TODO: WIP
-    def download_file_from_google_drive_using_api(self, sharing_link, join_id):
+    def _get_request_and_filename_from_drive_api(self, client, file_id):
+        """Get the request object and filename from a Google drive file_id. Will try to determine extension if missing.
+
+        Args:
+            client (pygsheets.Client): Authenticated client object from pyghseets
+            file_id (str): The Google fileId to be downloaded
+
+        Returns:
+            (googleapiclient.http.HttpRequest, str): Result of the .get_media() call, name of the file
+        """
+
+        get_media_request = client.drive.service.files().get_media(fileId=file_id)  # pylint:disable=no-member
+        metadata = client.drive.service.files().get(fileId=file_id).execute()  # pylint:disable=no-member
+        filename = metadata['name']
+        if not Path(filename).suffix:
+            try:
+                filename = filename + mimetypes.guess_extension(metadata['mimeType'])
+            except KeyError:
+                self._class_logger.warning('%s: No MIME type in drive info, file extension not set', file_id)
+            except TypeError:
+                self._class_logger.warning(
+                    '%s: Unable to determine file extension from MIME type, file extension not set', file_id
+                )
+
+        return get_media_request, filename
+
+    def _save_get_media_content(self, request, out_file_path):
+        """Save the binary data referenced from a .get_media() call to out_file_path
+
+        Args:
+            request (googleapiclient.http.HttpRequest): Result of the .get_media() call
+            out_file_path (Path): Path object to the location to save the data
+        """
+
+        in_memory = BytesIO()
+        downloader = MediaIoBaseDownload(in_memory, request)
+        done = False
+        while not done:
+            _, done = downloader.next_chunk()
+        out_file_path.write_bytes(in_memory.getbuffer())
+
+    def download_file_from_google_drive_using_api(self, gsheets_client, sharing_link, join_id):
+        """Download a file from Google Drive using an authenticated client and the Google API via pygsheets
+
+        Args:
+            gsheets_client (pygsheets.Client): The authenticated client object from pygsheets
+            sharing_link (str): Sharing link to the file to be downloaded
+            join_id (str or int): Unique key for the row (used for reporting)
+
+        Returns:
+            Path: Path of downloaded file or None if download fails/is not possible
+        """
         if not sharing_link:
             self._class_logger.debug('Row %s has no attachment info', join_id)
             return None
-        self._class_logger.debug('Row %s: downloading shared file %s', join_id, sharing_link)
+        self._class_logger.debug('Row %s: downloading file %s', join_id, sharing_link)
         try:
             file_id = self._get_file_id_from_sharing_link(sharing_link)
             self._class_logger.debug('Row %s: extracted file id %s', join_id, file_id)
-            response = self._get_http_response(file_id)
-            filename = self._get_filename_from_response(response)
+            get_media_request, filename = utils.retry(
+                self._get_request_and_filename_from_drive_api, gsheets_client, file_id
+            )
             out_file_path = self.out_dir / filename
-            self._save_response_content(response, out_file_path)
+            self._class_logger.debug('Row %s: writing to %s', join_id, out_file_path)
+            utils.retry(self._save_get_media_content, get_media_request, out_file_path)
             return out_file_path
         except Exception as err:
             self._class_logger.warning('Row %s: Couldn\'t download %s', join_id, sharing_link)
@@ -239,14 +302,42 @@ class GoogleDriveDownloader:
             dataframe (pd.DataFrame): Input dataframe with required columns
             sharing_link_column (str): Column holding the Google sharing link
             join_id_column (str): Column holding a unique key (for reporting purposes)
-            output_path_column (str): Column for the resulting path; will be added if it doesn't existing in the dataframe
+            output_path_column (str): Column for the resulting path; will be added if it doesn't existing in the
+            dataframe
 
         Returns:
             pd.DataFrame: Input dataframe with output path info
         """
 
         dataframe[output_path_column] = dataframe.apply(
-            lambda x: self.download_file_from_google_drive(x[sharing_link_column], x[join_id_column]), axis=1
+            lambda x: self.download_file_from_google_drive(x[sharing_link_column], x[join_id_column], pause=5), axis=1
+        )
+        return dataframe
+
+    def download_attachments_from_dataframe_using_api(
+        self, credentials, dataframe, sharing_link_column, join_id_column, output_path_column
+    ):
+        """Download the attachments linked in a dataframe column using an authenticated api client, creating a new
+        column with the resulting path
+
+        Args:
+            credentials (str or google.auth.credentials.Credentials): Path to the service file OR credentials object
+                obtained from google.auth.default() within a cloud function.
+            dataframe (pd.DataFrame): Input dataframe with required columns
+            sharing_link_column (str): Column holding the Google sharing link
+            join_id_column (str): Column holding a unique key (for reporting purposes)
+            output_path_column (str): Column for the resulting path; will be added if it doesn't existing in the
+            dataframe
+
+        Returns:
+            pd.DataFrame: Input dataframe with output path info
+        """
+
+        client = utils.authorize_pygsheets(credentials)
+
+        dataframe[output_path_column] = dataframe.apply(
+            lambda x: self.download_file_from_google_drive_using_api(client, x[sharing_link_column], x[join_id_column]),
+            axis=1
         )
         return dataframe
 
@@ -340,3 +431,43 @@ class SFTPLoader:
         if len(dataframe.index) == 0:
             self._class_logger.warning('Dataframe contains no rows. Shape: %s', dataframe.shape)
         return dataframe
+
+
+class PostgresLoader:
+    """Loads data from a Postgres/PostGIS database into a pandas data frame"""
+
+    def __init__(self, host, database, username, password, port=5432):
+        self._class_logger = logging.getLogger(__name__).getChild(self.__class__.__name__)
+
+        connection_string = f'postgresql://{username}:{password}@{host}:{port}/{database}'
+        self.engine = create_engine(connection_string)
+
+    def read_table_into_dataframe(self, table_name, index_column, crs, spatial_column):
+        """Read a table into a dataframe
+
+        Args:
+            table_name (str): Name of table or view to read in the following format: schema.table_name
+            index_column (str): Name of column to use as the dataframe's index
+            crs (str): Coordinate reference system of the table's geometry column
+            spatial_column (str): Name of the table's geometry or geography column
+
+        Returns:
+            pd.DataFrame.spatial: Table as a spatially enabled dataframe
+        """
+
+        self._class_logger.info('Reading `%s` into dataframe', table_name)
+
+        dataframe = gpd.read_postgis(
+            f'select * from {table_name}', self.engine, index_col=index_column, crs=crs, geom_col=spatial_column
+        )
+
+        spatial_dataframe = pd.DataFrame.spatial.from_geodataframe(dataframe, column_name=spatial_column)
+        for column in spatial_dataframe.select_dtypes(include=['datetime64[ns, UTC]']):
+            self._class_logger.debug('Converting column `%s` to ISO string format', column)
+            spatial_dataframe[column] = spatial_dataframe[column].apply(pd.Timestamp.isoformat)
+
+        self._class_logger.debug('Dataframe shape: %s', spatial_dataframe.shape)
+        if len(spatial_dataframe.index) == 0:
+            self._class_logger.warning('Dataframe contains no rows. Shape: %s', spatial_dataframe.shape)
+
+        return spatial_dataframe
